#include <NmraDcc.h>

// This Example shows how to use the library as a DCC Accessory Decoder or a DCC Signalling Decoder
// It responds to both the normal DCC Turnout Control packets and the newer DCC Signal Aspect packets 
// You can also print every DCC packet by uncommenting the "#define NOTIFY_DCC_MSG" line below

//151227 fujigay2 (12V - 24V 2ch point decoder version) 
//160505 Digitrax対応

#define MOTOR_A 9
#define MOTOR_B 10
#define MOTOR_C 8
#define MOTOR_D 7
#define LED 4
#define VOLT_READ A0
#define VOLT_WRITE 3

// CV Addresses
#define CV_ACCESSORY_OPERATION_TYPE 33
#define CV_ACCESSORY_OUTPUT_VOLTAGE_100MV 34
#define CV_ACCESSORY_OPERATION_TIME_10MS 35



const int DccAckPin = 9 ;// number 3 uses for VOLT_WRITE.

//VINというかコイルへの入力電圧に12Vを印加しているのが前提になります。電圧は変えてもOK（補正を自動でかけます）

//電圧指令値24V, 728d =(1024d/5V)*24V*(47k/317k)
//int gVref = 728;
//int gVref = 455;//15V
int gVref = 606;//20V
int gVref_cmd = 0;
int volt = 15;

NmraDcc  Dcc ;
DCC_MSG  Packet ;

struct CVPair
{
  uint16_t  CV;
  uint8_t   Value;
};
  
CVPair FactoryDefaultCVs [] =
{
  {CV_ACCESSORY_DECODER_ADDRESS_LSB, 3},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},
  {CV_29_CONFIG, 0b10000000},
  {CV_ACCESSORY_OPERATION_TYPE, 0},//個別
  {CV_ACCESSORY_OUTPUT_VOLTAGE_100MV, 150},//15[V]
  {CV_ACCESSORY_OPERATION_TIME_10MS, 5}//50[ms]

};

uint8_t FactoryDefaultCVIndex = 0;
unsigned long gPreviousWorkTime = 0;
uint16_t gAccessoryAddress = 1;

uint8_t  gCV33_OperationType = 0;
uint8_t  gCV34_OutputVoltage100mV = 150;
uint8_t  gCV35_OperationTime10ms = 50;

//Task Schedule
unsigned long gPreviousL5 = 0;
boolean first_flag = true;

void OutputStop();
uint16_t getMyAddr_Acc(void);

void notifyCVResetFactoryDefault()
{
  // Make FactoryDefaultCVIndex non-zero and equal to num CV's to be reset 
  // to flag to the loop() function that a reset to Factory Defaults needs to be done
  FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs)/sizeof(CVPair);
};

//const int DccAckPin = 3 ;

// This function is called by the NmraDcc library when a DCC ACK needs to be sent
// Calling this function should cause an increased 60ma current drain on the power supply for 6ms to ACK a CV Read 
void notifyCVAck(void)
{
  Serial.println("notifyCVAck") ;
  
  digitalWrite( DccAckPin, HIGH );
  delay( 6 );  
  digitalWrite( DccAckPin, LOW );
}

// Uncomment to print all DCC Packets
//#define NOTIFY_DCC_MSG
#ifdef  NOTIFY_DCC_MSG
void notifyDccMsg( DCC_MSG * Msg)
{
  Serial.print("notifyDccMsg: ") ;
  for(uint8_t i = 0; i < Msg->Size; i++)
  {
    Serial.print(Msg->Data[i], HEX);
    Serial.write(' ');
  }
  Serial.println();
}
#endif

// This function is called whenever a normal DCC Turnout Packet is received
void notifyDccAccState( uint16_t Addr, uint16_t BoardAddr, uint8_t OutputAddr, uint8_t State)
{
	uint16_t aAccAdress = Addr;

		
		//DEBUG
		Serial.print("notifyDccAccState: ") ;
		Serial.print(Addr,DEC) ;
		Serial.print(",BAdr:");
		Serial.print(BoardAddr,DEC) ;
		Serial.print(",OAdr:");
		Serial.print(OutputAddr,DEC) ;
		Serial.print(", Target Adr:");
		Serial.print(aAccAdress,DEC) ;
		Serial.print(",");
		Serial.println(State, HEX) ;


	//アドレスチェック(CVから得た11bit分の信号を比較)
  //gAccessoryAddress ,gAccessoryAddress + 1
	if((gCV33_OperationType == 0) && ( gAccessoryAddress != aAccAdress) && ( (gAccessoryAddress + 1) != aAccAdress))
	{
		return;
	}
 else if((gCV33_OperationType == 1) && ( gAccessoryAddress != aAccAdress))
  {
    return;
  }
	OutputStop();	
	
  uint8_t aDir = (OutputAddr & 0b001);


  digitalWrite(LED,HIGH);
  if((gCV33_OperationType == 0) && ( gAccessoryAddress == aAccAdress))
  {
    if(aDir  == 0)
    {
      Serial.println("Diverse") ;
    	digitalWrite(MOTOR_D, LOW);
    	digitalWrite(MOTOR_C, HIGH);
    }
    else if( aDir == 1)
    {
      Serial.println("Straight") ;
      digitalWrite(MOTOR_C, LOW);
      digitalWrite(MOTOR_D, HIGH);
    }
  }
  else if((gCV33_OperationType == 0) && ( (gAccessoryAddress + 1) == aAccAdress))
  {
    if(aDir  == 0)
    {
      Serial.println("Diverse") ;
      digitalWrite(MOTOR_B, LOW);
      digitalWrite(MOTOR_A, HIGH);
    }
    else if( aDir == 1)
    {
      Serial.println("Straight") ;
      digitalWrite(MOTOR_A, LOW);
      digitalWrite(MOTOR_B, HIGH);
    }
  }
  else if((gCV33_OperationType == 1) && ( gAccessoryAddress == aAccAdress))
  {
    if(aDir  == 0)
    {
      Serial.println("Diverse") ;
      digitalWrite(MOTOR_D, LOW);
      digitalWrite(MOTOR_C, HIGH);
      digitalWrite(MOTOR_B, LOW);
      digitalWrite(MOTOR_A, HIGH);
    }
    else if( aDir == 1)
    {
      Serial.println("Straight") ;
      digitalWrite(MOTOR_C, LOW);
      digitalWrite(MOTOR_D, HIGH);
      digitalWrite(MOTOR_A, LOW);
      digitalWrite(MOTOR_B, HIGH);
    }
  }

	//Wait XX * 10 [ms]
	delay((int)gCV35_OperationTime10ms * 10);
		
	//Stop output
	OutputStop();
  digitalWrite(LED,LOW);

  //Wait XX * 10 [ms]
  delay(550 - (int)gCV35_OperationTime10ms * 10);
  digitalWrite(LED,HIGH);
  delay(50);
  digitalWrite(LED,LOW);
}

uint16_t getMyAddr_Acc(void)
{
  uint16_t  Addr ;
  uint8_t   CV29Value ;

  CV29Value = Dcc.getCV( CV_29_CONFIG ) ;

  if( CV29Value & 0b10000000 )  // Accessory Decoder? 
    Addr = ( Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_MSB ) << 8 ) | Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_LSB ) ;

  else   // Multi-Function Decoder?
  {
    if( CV29Value & 0b00100000 )  // Two Byte Address?
      Addr = ( Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8 ) | Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB ) ;

    else
      Addr = Dcc.getCV( 1 ) ;
  }

  return Addr ;
}

// This function is called whenever a DCC Signal Aspect Packet is received
void notifyDccSigState( uint16_t Addr, uint8_t OutputIndex, uint8_t State)
{
  Serial.print("notifyDccSigState: ") ;
  Serial.print(Addr,DEC) ;
  Serial.print(',');
  Serial.print(OutputIndex,DEC) ;
  Serial.print(',');
  Serial.println(State, HEX) ;
}

void setup()
{
  Serial.begin(115200);
  Serial.println("Turnout decoder (12V - 24V)");
  
  // Configure the DCC CV Programing ACK pin for an output
  pinMode( DccAckPin, OUTPUT );


  //電圧制御ピン
  // D3を31kHzキャリア周期にする(コイルの定数を下げるため)
  TCCR2B = TCCR2B & B11111000 | B00000001;
  analogWrite(VOLT_WRITE, 0);//電圧 0 書き込み
  // Setup which External Interrupt, the Pin it's associated with that we're using and enable the Pull-Up 
  Dcc.pin(0, 2, 0);

  
	//出力ピン Motor_A,B,C,Dを出力にセット
	pinMode(MOTOR_A, OUTPUT);
	pinMode(MOTOR_B, OUTPUT);
  pinMode(MOTOR_C, OUTPUT);
  pinMode(MOTOR_D, OUTPUT);
  //出力を０リセット
	digitalWrite(MOTOR_A, 0);
	digitalWrite(MOTOR_B, 0);
  digitalWrite(MOTOR_C, 0);
  digitalWrite(MOTOR_D, 0);

  //LEDピン出力セット
  pinMode(LED, OUTPUT);
  digitalWrite(LED, LOW);

  
  // Call the main DCC Init function to enable the DCC Receiver
  Dcc.init( MAN_ID_DIY, 10, FLAGS_OUTPUT_ADDRESS_MODE | FLAGS_DCC_ACCESSORY_DECODER, 0 );
  gPreviousL5 =millis();
  
  //アクセサリアドレス(下位2bitを考慮)の先修得
  gAccessoryAddress = getMyAddr_Acc();
  Serial.print("Acc Adr is ");
  Serial.println(gAccessoryAddress);  
  

  Serial.println("Init Done");
}

void loop()
{

  // You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
  Dcc.process();
  
  if( FactoryDefaultCVIndex && Dcc.isSetCVReady())
  {
    FactoryDefaultCVIndex--; // Decrement first as initially it is the size of the array 
    Dcc.setCV( FactoryDefaultCVs[FactoryDefaultCVIndex].CV, FactoryDefaultCVs[FactoryDefaultCVIndex].Value);
  }

  //電圧制御
  if( (millis() - gPreviousL5) >= 33)
  {
    if(first_flag == true)
    {
      first_flag = false;
      //userCV のLoad
      gCV33_OperationType = Dcc.getCV(CV_ACCESSORY_OPERATION_TYPE);
      gCV34_OutputVoltage100mV = Dcc.getCV(CV_ACCESSORY_OUTPUT_VOLTAGE_100MV);
      gCV35_OperationTime10ms = Dcc.getCV(CV_ACCESSORY_OPERATION_TIME_10MS);
      //上限、下限の確認
      if(gCV34_OutputVoltage100mV > 240)//２４V以上 -> 24V
      {
        gCV34_OutputVoltage100mV = 240;
      }
      if(gCV35_OperationTime10ms > 50)//500ms以上 -> 500ms
      {
        gCV35_OperationTime10ms = 50;
      }
      Serial.print("CV_29_CONFIG (29) = ");
      Serial.println(Dcc.getCV( CV_29_CONFIG ));  
      Serial.print("CV_ACCESSORY_OPERATION_TYPE (33) = ");
      Serial.println(gCV33_OperationType);  
      Serial.print("CV_ACCESSORY_OUTPUT_VOLTAGE_100MV (34) = ");
      Serial.println(gCV34_OutputVoltage100mV);  
      Serial.print("CV_ACCESSORY_OPERATION_TIME_10MS (35) = ");
      Serial.println(gCV35_OperationTime10ms);  
    
      //value -> voltage
      gVref = (float)gCV34_OutputVoltage100mV  * (float)728 / (float)240;
    }
  	VRefCont();
    gPreviousL5 =millis();
  }
  
}

void OutputStop()
{
	digitalWrite(MOTOR_A, LOW);
	digitalWrite(MOTOR_B, LOW);
  digitalWrite(MOTOR_C, LOW);
  digitalWrite(MOTOR_D, LOW);
  
}

void VRefCont()
{
  //出力値検出
  int aVout = analogRead(VOLT_READ); 
  
  if( gVref > aVout)
  {
    gVref_cmd++;  
  }
  else
  {
    //gVref_cmd--; 
    gVref_cmd -= 10; 
  }
  if( gVref_cmd > 128)
  {
    gVref_cmd = 128;
  }

  if( gVref_cmd < 0)
  {
    gVref_cmd = 0;
  }
  //Log
/*
  Serial.print(", Vout="); 
  Serial.print(aVout);
  Serial.print(" ,PWM_Ref=");  
  Serial.println(gVref_cmd);  
*/
  //PWM出力
  analogWrite(VOLT_WRITE, gVref_cmd);
}



